{$I-,F+,O+,X+}
unit EditConv;

interface

uses FEditor;

const
  cvUpcase = 0;
  cvLocase = 1;
  cvLatin  = 2;
  cvRus    = 3;
  cvTable  = 4;
  cvFTable = 5;
  cvCaps   = 6;
  cvShort  = 7;
  cvLong   = 8;

procedure ConvertBlockPrim(Editor: PFileEditor; Action: Byte);
procedure ConvertToMoldovaPrim(Editor: PFileEditor);

implementation

uses Objects, Dialogs, App, Swaps, Keyboard, EditFile, EditMsg, LFN,
  TPString, Defaults;

procedure ConvertBlockPrim(Editor: PFileEditor; Action: Byte);
var
  B: Byte;
  Changed, UseLineSel: Boolean;
  C: Char;
  I, IB, IE: Integer;
  Pos: TPos;
  F: File;
  S, SubS: String;
  T: ConvertTable;
begin
  if (Action in [cvTable, cvFTable]) and
    not GetConvertTable(T, Action = cvFTable, Changed) then Exit;
  if Action in [cvLatin, cvRus] then
  begin
    for C := #0 to #255 do T[C] := C;
    if Action = cvLatin then
    begin
      Move(TByteArray(GetStrAddr(LoChars)^)[1], T[#128], Length(LoChars));
      Move(TByteArray(GetStrAddr(HiChars)^)[1], T[#224], Length(HiChars));
    end
    else Move(TByteArray(GetStrAddr(RuChars)^)[1], T[#34], Length(RuChars));
  end;
  Changed := False;
  UseLineSel := not (Action in [cvShort, cvLong]);
  with Editor^ do
  begin
    SaveLastLine;
    if IsSelected then
    begin
      IB := Area^.KB;
      if Area^.PK = 1 then IE := Area^.KK - 1
      else IE := Area^.KK;
    end
    else
    begin
      IB := Delta.Y + Cursor.Y;
      IE := IB;
    end;
    FastGetLine := True;
    for I := IB to IE do
    begin
      GetValidLine(I, S);
      if IsSelected and UseLineSel then
      begin
	GetPos(Area, I, Pos);
	SubS := System.Copy(S, Pos.AStart, Pos.AnEnd - Pos.AStart);
      end
      else SubS := S;
      if SubS <> '' then
      begin
	case Action of
	  cvUpcase: UpcaseSt(SubS);
	  cvLocase: LocaseSt(SubS);
	  cvCaps: for B := 1 to Length(SubS) do
	    begin
	      if B = 1 then
		if not IsSelected or (Pos.AStart = 1) then C := #0
		else C := S[Pos.AStart - 1]
	      else C := SubS[B - 1];
	      if (C = #0) or (C in LocalWordDelims) then
		SubS[B] := TPString.Upcase(SubS[B]);
	    end;
	  cvShort: lGetShortFileName(SubS, SubS);
	  cvLong: lGetLongFileName(SubS, SubS);
	  else for B := 1 to Length(SubS) do SubS[B] := T[SubS[B]];
	end;
	if IsSelected and UseLineSel then
	  Move(SubS[1], S[Pos.AStart], Length(SubS))
	else S := SubS;
	if not Changed and (UndoChain = nil) then
	begin
	  NewUndoChain(IE - IB + 1);
	  DrawStatusLine('Converting text characters.');
	end;
	ReplaceLine(I, S);
	Changed := True;
	if IsSelected and not MarkColumns and (I = Area^.KK) and
	  not UseLineSel then SetArea(
	    Area, Area^.KB, Area^.PB, Area^.KK, Length(S) + 1);
	if PSwapLinesCollection(FileLines)^.Status = coOverflow then
	begin
	  PSwapLinesCollection(FileLines)^.Status := 0;
	  Break;
	end;
      end;
      if (KbdStatus = 1) and AbortAction then Break;
    end;
    if not Changed then Exit;
    AddUndoChain;
    IsModified := True;
    DebugModified := True;
  end;
end;

function GetMoldovaStr(C, N: Char): String;
var
  S: String[4];
begin
  case C of
    ''     	: S := 'A';
    ''     	: S := 'B';
    ''     	: S := 'V';
    ''     	: S := 'G';
    ''		: S := 'D';
    '',''     : S := 'E';
    ''     	: if TPString.Upcase(N) = N then S := 'ZH' else S := 'Zh';
    ''     	: S := 'Z';
    ''     	: S := 'I';
    ''     	: if TPString.Upcase(N) = N then S := 'IJ' else S := 'Ij';
    ''     	: S := 'K';
    ''     	: S := 'L';
    ''     	: S := 'M';
    ''     	: S := 'N';
    ''     	: S := 'O';
    ''     	: S := 'P';
    ''     	: S := 'R';
    ''     	: S := 'S';
    ''     	: S := 'T';
    ''     	: S := 'U';
    ''     	: S := 'F';
    ''     	: if TPString.Upcase(N) = N then S := 'KH' else S := 'Kh';
    ''     	: if TPString.Upcase(N) = N then S := 'TS' else S := 'Ts';
    ''     	: if TPString.Upcase(N) = N then S := 'CH' else S := 'Ch';
    ''		: if TPString.Upcase(N) = N then S := 'SH' else S := 'Sh';
    ''	        : if TPString.Upcase(N) = N then S := 'SHCH' else S := 'Shch';
    ''         : S := '~';
    ''     	: S := 'Y';
    ''	        : S := '"';
    ''     	: if TPString.Upcase(N) = N then S := 'YU' else S := 'Yu';
    ''     	: if TPString.Upcase(N) = N then S := 'YA' else S := 'Ya';
    #240        : if TPString.Upcase(N) = N then S := 'JO' else S := 'Jo';
    ''     	: S := 'a';
    ''     	: S := 'b';
    ''     	: S := 'v';
    ''     	: S := 'g';
    ''     	: S := 'd';
    '',''     : S := 'e';
    ''     	: S := 'zh';
    ''     	: S := 'z';
    ''     	: S := 'i';
    ''     	: S := 'ij';
    ''     	: S := 'k';
    ''     	: S := 'l';
    ''     	: S := 'm';
    ''     	: S := 'n';
    ''     	: S := 'o';
    ''     	: S := 'p';
    ''     	: S := 'r';
    ''     	: S := 's';
    ''     	: S := 't';
    ''     	: S := 'u';
    ''     	: S := 'f';
    ''     	: S := 'kh';
    ''     	: S := 'ts';
    ''     	: S := 'ch';
    ''		: S := 'sh';
    ''		: S := 'shch';
    '' 	: S := '`';
    ''     	: S := 'y';
    ''	        : S := '''';
    ''     	: S := 'yu';
    ''     	: S := 'ya';
    #241        : S := 'jo';
    else S := C;
  end;
  GetMoldovaStr := S;
end;

procedure ConvertToMoldovaPrim(Editor: PFileEditor);
var
  Changed: Boolean;
  B, L: Byte;
  C, N: Char;
  I, IB, IE: Integer;
  Pos: TPos;
  SS: String[4];
  S: String;
begin
  Changed := False;
  with Editor^ do
  begin
    SaveLastLine;
    if IsSelected then
    begin
      IB := Area^.KB;
      if Area^.PK = 1 then IE := Area^.KK - 1
      else IE := Area^.KK;
    end
    else
    begin
      IB := Delta.Y + Cursor.Y;
      IE := IB;
    end;
    FastGetLine := True;
    for I := IB to IE do
    begin
      GetValidLine(I, S);
      L := Length(S);
      if IsSelected then GetPos(Area, I, Pos)
      else
      begin
	Pos.AStart := 1;
	Pos.AnEnd := L + 1;
      end;
      if Pos.AStart <= Length(S) then
      begin
	while (Pos.AStart <= Length(S)) and (Pos.AStart < Pos.AnEnd) do
	begin
	  C := S[Pos.AStart];
	  if Pos.AStart < Length(S) then N := S[Pos.AStart + 1]
	  else N := ' ';
	  SS := GetMoldovaStr(C, N);
	  B := Length(SS) - 1;
	  if B = 0 then S[Pos.AStart] := SS[1]
	  else
	  begin
	    System.Delete(S, Pos.AStart, 1);
	    System.Insert(SS, S, Pos.AStart);
	    Inc(Pos.AStart, B);
	    Inc(Pos.AnEnd, B);
	  end;
	  Inc(Pos.AStart);
	end;
	if not Changed and (UndoChain = nil) then
	begin
	  NewUndoChain(IE - IB + 1);
	  DrawStatusLine('Converting text to Volapyuk.');
	end;
	ReplaceLine(I, S);
	Changed := True;
	if IsSelected and not MarkColumns and (I = Area^.KK) then
	  SetArea(Area, Area^.KB, Area^.PB, Area^.KK,
	    Area^.PK + Length(S) - L);
	if PSwapLinesCollection(FileLines)^.Status = coOverflow then
	begin
	  PSwapLinesCollection(FileLines)^.Status := 0;
	  Break;
	end;
      end;
      if (KbdStatus = 1) and AbortAction then Break;
    end;
    if not Changed then Exit;
    AddUndoChain;
    IsModified := True;
    DebugModified := True;
  end;
end;

end.
